/*
 * pop3-gw.c, a Post Office Protocol version 3 proxy designed to work
 *                with TIS fwtk
 *
 * (C) Copyright 1998 by ArkanoiD <ark@eltex.ru>
 * Version 0.3alpha
 *
 */

#include <stdio.h>
#include <stdlib.h>  
#include <string.h>
#include <syslog.h>
#include <signal.h>
#include <unistd.h> 
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <md5.h> 
#include <net/if.h>
#include "firewall.h"

#ifdef IPFILTER
#include "ip_compat.h"
#include "ip_fil.h"
#include "ip_nat.h"
#endif

extern h_errno;

typedef struct {
	int	inbytes;
	int	outbytes;
} iostats;

#define OP_DENY 001
#define OP_LOG  002
#define OP_AUTH 004		/* never used here */
#define OP_XTND 010

typedef	struct {
	char	*name;
	int	flg;
	int	mline;
} pop3_action;

/*
 * Only transaction state commands are subject of extended control
 *
 */

static	pop3_action	tstate_actions[] = {
	"STAT",		0,	0,
	"LIST",		0,	1,
	"RETR",		0,	1,
	"DELE",		0,	0,
	"NOOP",		0,	0,
	"RSET",		0,	0,
	"UIDL",		0,	1,
	"TOP",		0,	1,
	0,		0,	0
};

static	char	rladdr[512];
static	char	riaddr[512];

static	char	plug_server[512] ="";

static	time_t	ontime;
static	time_t	offtime;
static	int	timeout = PROXY_TIMEOUT;
static	int	fallback = 0;
static	int	transparent = 0;
static	int	apop_only = 0;
static	int	server_port = 110;
static	int	extendperm = 0;

static	iostats	iodata;

static	Cfg	*confp;
static	char	**validests = (char **)0;
static	char	**validusers = (char **)0;
static	char	authuser[128] = "unauth";
static  char	separator[2] = "@";


void proxy_exit(dest)
char	*dest;
{
	time(&offtime);
	syslog(LLEV,"exit host=%.512s/%.20s dest=%.128s in=%d out=%d duration=%d",
                     rladdr,riaddr,dest,iodata.inbytes,iodata.outbytes,
		     (int)offtime - ontime);
	exit(1);
}

oktotalkto()
{                       
        Cfg     *cf;
        int     x;
   
        cf = cfg_get("hosts",confp);
        while(cf != (Cfg *)0) {
                if(cf->argc < 1)
                        goto skip;

                for(x = 0; x < cf->argc && cf->argv[x][0] != '-'; x++) {
                        if(hostmatch(cf->argv[x],riaddr)) {
                                if(cf->flags & PERM_DENY) {
                                        syslog(LLEV,"deny host=%.512s/%.20s use of gateway",rladdr,riaddr);  
                                        exit(1);
                                }
                                syslog(LLEV,"permit host=%.512s/%.20s use of gateway",rladdr,riaddr);
                                return(acceptrule(cf) == 0);
                        }
                }
skip:
                cf = cfg_get("hosts",(Cfg*)0);
        } 
        syslog(LLEV,"deny host=%.512s/%.20s use of gateway",rladdr,riaddr);
        exit(1);
}

static  void                    
accept_setdest(v,c)
char    *v;
int     c;      
{                       
        int     dests = 0;

        if(validests == (char **)0)
                validests = (char **)malloc(sizeof(char *) * 2);
        else {
                for(dests = 0; validests[dests] != (char *)0; dests++)
                        ;
                validests = (char **)realloc(validests,(dests + 2) * sizeof(char *));
        }
        if(validests == (char **)0)
                return;
        validests[dests] = v;   
        validests[dests + 1] = (char *)0;
}

static  void                    
accept_setuser(v,c)
char    *v;
int     c;      
{                       
        int     users = 0;

        if(validusers == (char **)0)
                validusers = (char **)malloc(sizeof(char *) * 2);
        else {
                for(users = 0; validusers[users] != (char *)0; users++)
                        ;
                validusers = (char **)realloc(validusers,(users + 2) * sizeof(char *));
        }
        if(validusers == (char **)0)
                return;
        validusers[users] = v;   
        validusers[users + 1] = (char *)0;
}

static  void                    
accept_setlog(v,c)
char    *v;
int     c;      
{ 
	pop3_action	*act;

	for (act = tstate_actions; act->name != (char *)0; act++)
		if(!strcasecmp(v,act->name)) {
			act->flg |= OP_LOG;
			return;
		}
	if (act->name == (char *)0)
		syslog(LLEV,"fwtkcfgerr: config line %d: bad operand %.100s",c,v);
}
	
static  void                    
accept_setdeny(v,c)
char    *v;
int     c;      
{ 
	pop3_action	*act;

	for (act = tstate_actions; act->name != (char *)0; act++)
		if(!strcasecmp(v,act->name)) {
			act->flg |= OP_DENY;
			return;
		}
	if (act->name == (char *)0)
		syslog(LLEV,"fwtkcfgerr: config line %d: bad operand %.100s",c,v);
}

acceptrule(c)
Cfg     *c;
{
        int     x;
        void    (*op)();
 
        for(x = 1; x < c->argc; x++) {
                if(c->argv[x][0] != '-')
                        continue;

		if (!strcmp(c->argv[x], "-plug-to")) {
                        x++;
                        if (x >= c->argc) {
                                syslog(LLEV,"fwtkcfgerr: -plug-to needs an argument, line %d",c->ln);
                                exit(1);
                        }
                        strncpy(plug_server, c->argv[x], strlen(c->argv[x]));
                        plug_server[strlen(c->argv[x])] = '\0';
                        continue;
                }

		if (!strcmp(c->argv[x], "-authuser")) {
                        x++;
                        if (x >= c->argc) {
                                syslog(LLEV,"fwtkcfgerr: -authuser needs an argument, line %d",c->ln);
                                exit(1);
                        }
                        strncpy(authuser, c->argv[x], strlen(c->argv[x]));
                        authuser[strlen(c->argv[x])] = '\0';
                        continue;
                }

		if (!strcmp(c->argv[x], "-separator")) {
                        x++;
                        if (x >= c->argc) {
                                syslog(LLEV,"fwtkcfgerr: -separator needs an argument, line %d",c->ln);
                                exit(1);
                        }
			if (strlen(c->argv[x]) == 1)
				separator[0] = c->argv[x][0];
			else {
                                syslog(LLEV,"fwtkcfgerr: separator can be one character only, line %d",c->ln);
                                exit(1);
                        }
                        continue;
		}

		if(!strcmp(c->argv[x],"-fallback")) {
			fallback = 1;
			continue;
		}

		if(!strcmp(c->argv[x],"-transparent")) {
			transparent = 1;
			continue;
		}

		if(!strcmp(c->argv[x],"-apop-only")) {
			apop_only = 1;
			continue;
		}

		if(!strcmp(c->argv[x],"-extnd")) {
			extendperm = 1;
			continue;
		}

		/* options that take parameters and lists */


		op = 0;
		if(!strcmp(c->argv[x],"-dest"))
			op = accept_setdest;
		if(!strcmp(c->argv[x],"-user"))
			op = accept_setuser;
		if(op == 0) {
			syslog(LLEV,"fwtkcfgerr: bad option line %d: %.100s",c->ln,c->argv[x]);
			exit(1);
		}
		if(++x >= c->argc) {
			syslog(LLEV,"fwtkcfgerr: malformed line %d: missing option",c->ln);
			exit(1);
		}
                if(!strcmp(c->argv[x],"{")) {
                        while(1) {
                                if(++x >= c->argc) {
                                        syslog(LLEV,"fwtkcfgerr: malformed line %d: missing option",c->ln);
                                        exit(1);
                                }
                                if(!strcmp(c->argv[x],"}"))
                                        break;
                                (*op)(c->argv[x],c->ln);
                        }
                } else
                        (*op)(c->argv[x],c->ln);

        }
	return(0);
}

check_dest(host)
char	*host;
{
	char	buffer[1024];

	if(validests != (char **)0) {
		char    **xp;
		int     x;

		for(xp = validests; *xp != (char *)0; xp++) {
			if(**xp == '!' && hostmatch(*xp + 1,host)) {
				syslog(LLEV,"deny host=%.512s/%.20s connect to %.512s",rladdr,riaddr,host);
				return(1);
			} else if(hostmatch(*xp,host))
				break;
		}
		if(*xp == (char *)0) {
			syslog(LLEV,"deny host=%.512s/%.20s connect to %.512s",rladdr,riaddr,host);
			return(1);
		}
	}
	if(extendperm) {
		switch (auth_perm(confp, authuser, "pop3-gw", host,(char *)0)) {
		case -1:
			syslog(LLEV,"no match for pop3-gw user=%.100s in netperm-table",authuser);
			return(1);
		case 1: 
			syslog(LLEV,"deny host=%.512s/%.20s connect to %.512s user=%.100s",rladdr,riaddr,host,authuser);
			return(1);
		}
	}
	syslog(LLEV,"permit host=%.512s/%.20s connect to %.512s/%d",rladdr,riaddr,host,server_port);
	return(0);
}

check_user(user)
char	*user;
{
	char	buffer[1024];

	if(validusers != (char **)0) {
		char    **xp;
		int     x;

		for(xp = validusers; *xp != (char *)0; xp++) {
			if(**xp == '!' && !strcmp(*xp + 1,user)) {
				return(1);
			} else if(!strcmp(*xp,user) || !strcmp(*xp,"*"))
				break;
		}
		if(*xp == (char *)0) {
			return(1);
		}
	}
	return(0);
}

get_transparent_dst(dsthost)
char	*dsthost;
{
	int			len;
	struct sockaddr_in	to_addr;
#ifdef IPFILTER
	struct sockaddr_in	from_addr;
	int			ipfnat_fd;
	struct	natlookup	natlookup;
#endif

	bzero((char *)&to_addr, sizeof(to_addr));

	len = sizeof(to_addr);
	if (getsockname(0, (struct sockaddr *)&to_addr, &len) < 0) {
		syslog(LLEV,"fwtksyserr: getsockname() failed: %m");
		exit(1);
	}

#ifndef IPFILTER

	strncpy(dsthost,inet_ntoa(to_addr.sin_addr),sizeof(dsthost));
	return(0);

#else


	bzero((char *)&from_addr, sizeof(from_addr));
	if (getpeername(0, (struct sockaddr *)&from_addr, &len) < 0) {
		syslog(LLEV,"fwtksyserr: getpeername() failed: %m");
		exit(1);
	}

	natlookup.nl_inport = to_addr.sin_port;
	natlookup.nl_outport = from_addr.sin_port;
	natlookup.nl_inip = to_addr.sin_addr;
	natlookup.nl_outip = from_addr.sin_addr;
	natlookup.nl_flags = IPN_TCP;

	if ((ipfnat_fd = open(IPL_NAT, O_RDONLY)) < 0) {
		syslog(LLEV,"fwtksyserr: cannot open NAT deivce: %m");
		exit(1);
	}

	if (ioctl(ipfnat_fd, SIOCGNATL, &natlookup) == -1) {
		syslog(LLEV,"fwtksyserr: SIOCGNATL failed: %m");
		close(ipfnat_fd);
		exit(1);
	}

	strcpy(dsthost,inet_ntoa(natlookup.nl_realip));
	server_port = ntohs(natlookup.nl_realport);
	return(0);

#endif

}

get_multiline(int server_fd)
{
	char		buf[1024];
	int		x;

	while(1) {
		if(dgets(buf, 512, server_fd,(char *)0) <= 0) {
			syslog(LLEV,"error reading server data");
			return(1);
		}
		if (write(0,buf,strlen(buf)) != strlen(buf)) {
			syslog(LLEV,"error sending data to client");	
			return(1);
		}
		iodata.inbytes += strlen(buf);
		if (!strcmp(buf,".\r\n"))
			return(0);
	}
}

generic_tunnel(int sock1, int sock2)
{
	int		maxfd;
	fd_set		rset;
	char		buffer[16384 + 1];
	int		x;
	struct timeval	timo;

	if(sock1 > sock2)
		maxfd = sock1 + 1;
	else
		maxfd = sock2 + 1;

	while(1)
	{
		FD_ZERO(&rset);
		FD_SET(sock1,&rset);
		FD_SET(sock2,&rset);
		timo.tv_sec = timeout;
		timo.tv_usec = 0;
		if((x = select(maxfd, &rset, NULL, NULL, &timo)) < 0) {
			if (errno != EBADF && errno != EINTR)
				syslog(LLEV,"fwtksyserr: select: %m");
			exit(1);
		} else {
			if(x == 0) {
				syslog(LLEV,"Network timeout signal after %d seconds", timeout);
				break;
			}
		}

		if (FD_ISSET(sock1,&rset))
		{
			if((x = read(sock1,buffer,sizeof(buffer))) <= 0)
			{
				break;
			}
			if (write(sock2,buffer,x)!=x) 
				break;
			iodata.outbytes+=x;
		}
		
		if (FD_ISSET(sock2,&rset))
		{
			if((x = read(sock2,buffer,sizeof(buffer))) <= 0)
			{
				break;
			}
			if (write(sock1,buffer,x)!=x)
				break;
			iodata.inbytes+=x;
		}
	}
	close(sock1);
	close(sock2);
	return(0);
}

parse_user_cmd(buffer,username,server_host)
char	*buffer;
char	*username;
char	*server_host;
{
	char	*p;

	buffer[strlen(buffer)-2] = '\0';
	if ((p = strtok(buffer+5,separator)) != NULL)
		strncpy(username,p,32);
	if ((p = strtok(NULL,":")) != NULL)
		strncpy(server_host,p,64);
	if ((p = strtok(NULL,"\r\n")) != NULL)
		server_port = atoi(p);
	if (!server_port)
		server_port = str_to_port("pop3");
	if ((strlen(username)==0) || (strlen(server_host)==0))
		return(1);
	return(0);
}

parse_apop_cmd(buffer,username,digest)
char	*buffer;
char	*username;
char	*digest;
{
	char	*p;
	buffer[strlen(buffer)-2] = '\0';
	if ((p = strtok(buffer+5," ")) != NULL)
		strncpy(username,p,32);
	if ((p = strtok(NULL,"\r\n")) != NULL)
		strncpy(digest,p,32);
	return ((strlen(username)==0) || (strlen(digest)!=32));
}

parse_server_greeting(buffer,timestamp)
char	*buffer;
char	*timestamp;
{
	char	*p;

	buffer[strlen(buffer)-2] = '\0';

	if (strncasecmp(buffer,"+OK",3)) 
		return(1);
	
	strcpy(timestamp,"none");

	if (p = strchr(buffer,'<')) {
		if(!strchr(p,'>')) {
			syslog(LLEV,"warning: server provided no timestamp and probably does not support APOP");
		} else {
			strncpy(timestamp,p,64);
			syslog(LLEV,"received timestamp from server: %.64s",timestamp);
		}
	} else
		syslog(LLEV,"warning: server provided no timestamp and probably does not support APOP");
	return(0);

}

do_server_auth_raw(server_fd,username,digest)
int	server_fd;
char	*username;
char	*digest;
{
	char buf[512];

	sprintf(buf,"APOP %s %s\r\n",username,digest);
	write(server_fd,buf,strlen(buf));
	if(dgets(buf, 512, server_fd,(char *)0) <= 0) {
		syslog(LLEV,"error reading server data");
		return(1);
	}
	if (strncasecmp(buf,"+OK",3)) 
		return(1);

	sprintf(buf,"+OK authenticated to server via APOP\r\n");
	write(0,buf,strlen(buf));	
	return(0);
}

tohex(hex,raw,len)
char	*hex; 
char	*raw;
int	len;
{
	char *hdig="0123456789abcdef";

	while(len > 0){
		len--;
		*hex++ = hdig[((*raw)>>4)&0xf];
		*hex++ = hdig[(*raw++)&0xf];
	}
	*hex++ = '\0';
	return;
}

do_server_auth(server_fd,username,password,timestamp)
int	server_fd;
char	*username;
char	*password;
char	*timestamp;
{
	MD5_CTX	ctx;
	char digest_raw[17];
	char digest_hex[35];
	char buf[512];
	char *spec;

	password[strlen(password)-2]='\0';
	MD5Init(&ctx);
	MD5Update(&ctx,timestamp,strlen(timestamp));
	MD5Update(&ctx,password,strlen(password));
	MD5Final(digest_raw,&ctx);
	tohex(digest_hex,digest_raw,16);
	sprintf(buf,"APOP %s %s\r\n",username,digest_hex);
	write(server_fd,buf,strlen(buf));
	if(dgets(buf, 512, server_fd, spec) <= 0) {
		syslog(LLEV,"error reading server data");
		return(1);
	}
	if (strncasecmp(buf,"+OK",3)) 
		return(1);

	sprintf(buf,"+OK authenticated to server via APOP\r\n");
	write(0,buf,strlen(buf));	
	return(0);

}

do_server_pauth(server_fd,username,password)
int	server_fd;
char	*username;
char	*password;
{
	char buf[512];
	char *spec;

	sprintf(buf,"USER %s\r\n",username);
	write(server_fd,buf,strlen(buf));
	if(dgets(buf, 512, server_fd, (char *)0) <= 0) {
		syslog(LLEV,"error reading server data");
		return(1);
	}
	if (strncasecmp(buf,"+OK",3)) 
		return(1);
	sprintf(buf,"PASS %s\r\n",password);
	write(server_fd,buf,strlen(buf));
	if(dgets(buf, 512, server_fd, (char *)0) <= 0) {
		syslog(LLEV,"error reading server data");
		return(1);
	}
	if (strncasecmp(buf,"+OK",3)) 
		return(1);

	sprintf(buf,"+OK authenticated to server via USER/PASS\r\n");
	write(0,buf,strlen(buf));	
	return(0);
	
}


main(int argc, char *argv[])
{
	Cfg			*cf;
	char			buffer[1024];
	char			username[32];
	char			pass[512];
	char			timestamp[64];
	char			digest_hex[35];
	char			server_host[512];
	int			server_fd = -1;
	char			*spec;
	int			stage_complete;
	char			our_host[MAXHOSTNAMELEN];
	int			runuid = -1;
	int			rungid = -1;
	int			len;
	struct	sockaddr_in	to_addr;

#ifndef LOG_DAEMON
	openlog("pop3-gw",LOG_PID);
#else
	openlog("pop3-gw",LOG_PID|LOG_NDELAY,LFAC);
#endif

	if((confp = cfg_read("pop3-gw")) == (Cfg *)-1) 
		exit(1);

	if((cf = cfg_get("groupid",confp)) != (Cfg *)0) {

		if(cf->argc != 1) {
			syslog(LLEV,"fwtkcfgerr: groupid must have one parameter, line %d",cf->ln);
			exit(1);
		}
		if((rungid = mapgid(cf->argv[0])) == -1) {
			syslog(LLEV,"fwtkcfgerr: cannot map %.100s to gid",cf->argv[0]);
			exit(1);
		}
	}

	if((cf = cfg_get("userid",confp)) != (Cfg *)0) {

		if(cf->argc != 1) {
			syslog(LLEV,"fwtkcfgerr: userpid must have one parameter, line %d",cf->ln);
			exit(1);
		}
		if((runuid = mapuid(cf->argv[0])) == -1) {
			syslog(LLEV,"fwtkcfgerr: cannot map %.100s to uid",cf->argv[0]);
			exit(1);
		}
	}

	if(rungid != -1 && setgid(rungid)) {
		syslog(LLEV,"fwtksyserr: cannot setgid %d: %m",rungid);
		exit(1);
	}

	if(runuid != -1 && setuid(runuid)) {
		syslog(LLEV,"fwtksyserr: cannot setuid %d: %m",runuid);
		exit(1);
	}

	if (peername(0,rladdr,riaddr,sizeof(rladdr))!= 0) {
		syslog (LLEV, "fwtksyserr: cannot get peer name: %m");
		exit(1);
	}

	if((cf = cfg_get("timeout",confp)) != (Cfg *)0) {
		if(cf->argc != 1) {
			syslog(LLEV,"fwtkcfgerr: timeout must have one parameter, line %d",cf->ln);
			exit(1);
		}
		if((timeout = atoi(cf->argv[0])) <= 0) {
			syslog(LLEV,"fwtkcfgerr: timeout %.100s invalid, line %d",cf->argv[0],cf->ln);
			exit(1);
		}
	} else
		timeout = PROXY_TIMEOUT;

	if(!oktotalkto())
		exit(1);


	time(&ontime);
	iodata.inbytes  = 0;
	iodata.outbytes = 0;
	dgets_timeout(timeout);
	memset(buffer, 0, 1023);

	if ((*plug_server==*"") && !transparent) {
		gethostname(our_host,MAXHOSTNAMELEN);
		sprintf(timestamp,"<%d.%u@%.64s>", getpid(), ontime, our_host);
	} else {
		if (transparent) {
			get_transparent_dst(server_host);

			if (check_dest(server_host)) {
				sprintf(buffer,"-ERR POP3 gateway denied destination access\r\n");
				write (0, buffer, strlen(buffer));
				proxy_exit("none");
			}
		} else
			strncpy(server_host,plug_server,sizeof(server_host));

		if((server_fd = conn_server(server_host,server_port,0,(char *)0)) < 0) {
			syslog(LLEV,"cannot connect to server %s/%d: %m", server_host,server_port);
			sprintf(buffer,"-ERR POP3 gateway destination connect failed\r\n");
			write (0, buffer, strlen(buffer));
			proxy_exit("none");
		}
		if(dgets(buffer, 512, server_fd, spec) <= 0) {
			close(server_fd);
			syslog(LLEV,"error reading server data");
			sprintf(buffer,"-ERR POP3 gateway destination error\r\n");
			write (0, buffer, strlen(buffer));
			proxy_exit(server_host);
		}
		if (parse_server_greeting(buffer,timestamp) != 0) {
			close(server_fd);
			syslog(LLEV,"server greeting message is invalid");
			sprintf(buffer,"-ERR server greetimg message is invalid\r\n");
			write (0, buffer, strlen(buffer));
			proxy_exit(server_host);
		} 
	}

	sprintf(buffer,"+OK POP3 gateway ready %.128s\r\n",timestamp);
	write(0,buffer,strlen(buffer));

	while(!stage_complete) {
		if(dgets(buffer, 512, 0, spec) <= 0)
			proxy_exit("none");
		if (!strncasecmp(buffer,"USER ",5)) {
			if (apop_only) {
				sprintf(buffer,"-ERR insecure protocol is not permitted\r\n");
				write (0, buffer, strlen(buffer));
				goto skip_user;
			}
			if ((*plug_server==*"") && !transparent) {
				if (parse_user_cmd(buffer,username,server_host,server_port) != 0) {
					sprintf(buffer,"-ERR incorrect syntax, use user@host[:port] to specify destination\r\n"); 
					write (0, buffer, strlen(buffer));
				} else {
					if (check_user(username)) {
						sprintf(buffer,"-ERR access denied\r\n");	
						write (0, buffer, strlen(buffer));
                                        	syslog(LLEV,"deny host=%.512s/%.20s access for user %s",rladdr,riaddr,username);  
						proxy_exit("none");
					}
					if (check_dest(server_host)) {
						sprintf(buffer,"-ERR destination access denied\r\n");
						write (0, buffer, strlen(buffer));
						proxy_exit("none");
					}
	
					if((server_fd = conn_server(server_host,server_port,0,(char *)0)) < 0) {
						syslog(LLEV,"cannot connect to server %s/%d: %m", server_host,server_port);
						sprintf(buffer,"-ERR cannot connect to server\r\n");
						write (0, buffer, strlen(buffer));
						proxy_exit("none");
					}
					if(dgets(buffer, 512, server_fd, spec) <= 0) {
						close(server_fd);
						syslog(LLEV,"error reading server data");
						sprintf(buffer,"-ERR error reading server data\r\n");
						write (0, buffer, strlen(buffer));
						proxy_exit(server_host);
					}
					if (parse_server_greeting(buffer,timestamp) != 0) {
						close(server_fd);
						syslog(LLEV,"server greeting message is invalid");
						sprintf(buffer,"-ERR server greetimg message is invalid\r\n");
						write (0, buffer, strlen(buffer));
						proxy_exit(server_host);
					} 
					if (!strcmp(timestamp,"none") && !fallback) {
						close(server_fd);
						syslog(LLEV,"fallback to USER/PASS authentication is not permitted");
						sprintf(buffer,"-ERR server does not support APOP\r\n");
						write (0, buffer, strlen(buffer));
					} else 
						stage_complete = 1;
				}  
			} else {
				buffer[strlen(buffer)-2] = '\0';
				if ((buffer+5) && strlen(buffer+5)) {
					strncpy(username,buffer+5,sizeof(username));
					if (check_user(username)) {
						sprintf(buffer,"-ERR access denied\r\n");	
						write (0, buffer, strlen(buffer));
                                        	syslog(LLEV,"deny host=%.512s/%.20s access for user %s",rladdr,riaddr,username);  
						proxy_exit("none");
					}
					stage_complete = 1;
				} else {
					sprintf(buffer,"-ERR syntax error\r\n");
					write (0, buffer, strlen(buffer));
				}
			}

skip_user:

		} else if (!strncasecmp(buffer,"APOP ",5)) { 
			if (!strcmp(timestamp,"none") || ((*plug_server==*"") && !transparent)) {
				sprintf(buffer,"-ERR APOP is unsupported on client side\r\n");	
				write (0, buffer, strlen(buffer));
			} else {
				if (parse_apop_cmd(buffer,username,digest_hex)) {
					sprintf(buffer,"-ERR syntax error\r\n");
					write (0, buffer, strlen(buffer));
				} else {
					if (check_user(username)) {
						sprintf(buffer,"-ERR access denied\r\n");	
						write (0, buffer, strlen(buffer));
                                        	syslog(LLEV,"deny host=%.512s/%.20s access for user %s",rladdr,riaddr,username);  
						proxy_exit("none");
					}
					if (do_server_auth_raw(server_fd,username,digest_hex)) {
						sprintf(buffer,"-ERR APOP authentication failed\r\n");
						write (0, buffer, strlen(buffer));
						proxy_exit(server_host);
					} else goto transaction_state;
				}
			}
						

		} else if (!strncasecmp(buffer,"QUIT",4)) { 
			sprintf(buffer,"+OK proxy session terminated\r\n");	
			write (0, buffer, strlen(buffer));
			proxy_exit(server_host);

		} else {
			sprintf(buffer,"-ERR invalid command\r\n");
			write (0, buffer, strlen(buffer));
		}
	}

	sprintf(buffer,"+OK we are connected to server\r\n");
	write (0, buffer, strlen(buffer));

	stage_complete = 0;

	while (!stage_complete) {
		if(dgets(buffer, 512, 0, spec) <= 0) {
			syslog(LLEV,"error reading client data");
			proxy_exit(server_host);
		} else if (!strncasecmp(buffer,"QUIT",4)) { 
			sprintf(buffer,"+OK proxy session terminated\r\n");	
			write (0, buffer, strlen(buffer));
			proxy_exit(server_host);
		} else if (!strncasecmp(buffer,"PASS ",5)) { 
			strncpy(pass,buffer+5,sizeof(pass));
			stage_complete = 1;
		} else {
			sprintf(buffer,"-ERR PASS command required to complete AUTHORIZATION state\r\n");
			write (0, buffer, strlen(buffer));
		} 
	}

	if ( !strcmp(timestamp,"none") ||
	      (do_server_auth(server_fd,username,pass,timestamp) != 0)) {
/*
 *		APOP authentication failed or not available
 */
		if (!fallback) {
			close(server_fd);
			syslog(LLEV,"user failed to authenticate to server");
			sprintf(buffer,"-ERR server authentication failed\r\n");
			write (0, buffer, strlen(buffer));
			proxy_exit(server_host);
		} else if (strcmp(timestamp,"none")) {
/*
 *		APOP authentication failed so we have to reconnect to server
 */
			close(server_fd);
			syslog(LLEV,"warning: APOP authentication failed, reverting to insecure mode"); 
			if((server_fd = conn_server(server_host,server_port,0,(char *)0)) < 0) {
				syslog(LLEV,"cannot connect to server %s/%d: %m", server_host,server_port);
				sprintf(buffer,"-ERR cannot connect to server\r\n");
				write (0, buffer, strlen(buffer));
				proxy_exit("none");
			}
			if(dgets(buffer, 512, server_fd, spec) <= 0) {
				close(server_fd);
				syslog(LLEV,"error reading server data");
				sprintf(buffer,"-ERR error reading server data\r\n");
				write (0, buffer, strlen(buffer));
				proxy_exit(server_host);
			}
			if (strncasecmp(buffer,"+OK",3)) {
				close(server_fd);
				syslog(LLEV,"server greeting message is invalid");
				sprintf(buffer,"-ERR server greetimg message is invalid\r\n");
				write (0, buffer, strlen(buffer));
				proxy_exit(server_host);
			}
		}
/*
 *		Now trying USER/PASS
 */
	        if (do_server_pauth(server_fd,username,pass) != 0) {
			close(server_fd);
			syslog(LLEV,"user failed to authenticate to server");
			sprintf(buffer,"-ERR server authentication failed\r\n");
			write (0, buffer, strlen(buffer));
			proxy_exit(server_host);
		}
	}

/*
 *	Proceed to TRANSACTION state
 */
transaction_state:

	generic_tunnel(0,server_fd);
	proxy_exit(server_host);

}


